\ ----------------------------------------------------------------------
\ LANGUAGE    : AVR ByteForth vsn 2.00
\ PROJECT     : Round Robin multitasker voor AVR
\ DESCRIPTION : Op zoek naar multitasking voor ByteForth
\ CATEGORY    : Applicatie, afmeting: 122 bytes.
\ AUTHOR      : Willem Ouwerkerk, April 27, 2002
\ LAST CHANGE : Willem Ouwerkerk, May 20, 2002
\ ----------------------------------------------------------------------

\ 90S8515                                 \ Selecteer target
\ NEEDS TARGET.FRT                        \ Laad cpu specifiek deel


DOC umentatie van een Round Robin scheduler

Deze implementatie is een stap in de zoektocht naar een
efficiente en elegante multitasker voor (AVR) ByteForth.
Er wordt voor elke taak een aparte return- en data-stack 
gemaakt.

De data-stack wordt eerst verlaagd en dan beschreven, de 
return-stack wordt eerst beschreven en dan pas verlaagd 
(subtiel verschil). De lage byte wordt bij een call eerst 
op de return-stack geplaatst.

Er worden voor stacks en pointer opslag 28 variabelen per 
extra taak gebruikt, 1 register en 122 bytes code. Er is geen 
statusvlag voor de taken ingebouwd, een taak blijft daarom 
na activering altijd doorlopen.


Er zijn 2 belangrijke woorden, de namen van deze woorden 
zijn afkomstig van functioneel identieke woorden in een 
artikel uit FD volume 5 nr. 4 blz: 27 van Henry Laxen.

PAUSE       Bewaar RP, SP, naar volgende taak, zet RP van 
            die taak, zet SP van die taak en ga via RET, die 
            taak uitvoeren.

ACTIVATE    Behalve taak 0 moeten alle andere taken geactiveerd 
            worden; SP wordt klaargezet, het XT van de taak
            wordt op zijn return-stack geplaatst. De RP wordt
            op de data-stack bewaard.

MS1 MS2 MS3 Drie afzonderlijke MS routines voor drie 
            afzonderlijke taken. Ze lopen onafhankelijk van
            elkaar m.b.v. een timer-interrupt.

ENDDOC


04 CONSTANT #TASKS              \ Aantal gewenste taken
12 CONSTANT STACK-SPACE         \ Datastacks zijn 10 bytes diep
20 CONSTANT RETURNSTACK-SPACE   \ Returnstacks zijn 8 cellen diep

REGISTER TASK#                  \ Bevat het nummer van de actieve taak
#TASKS 2VARIABLES TASKS         \ Datastack pointers van taken in slaap

#TASKS 1- STACK-SPACE *         \ Calculate
VARIABLES STACKS                \ RAM voor datastacks
#TASKS 1- RETURNSTACK-SPACE *   \ Calculate
VARIABLES RETURNSTACKS          \ RAM voor returnstacks


\ Start van de code routines

SPL SFR LRP                     \ De echte cpu registers
SPH SFR HRP
26  (REGISTER) LSP              \ 26 = X-register low
27  (REGISTER) HSP

\ Naar volgende taak 48 cycles = 12 usec. bij een 4 MHz xt.
CODE PAUSE      ( -- )
    R1 PUSH,                    \ Save loop register
    ZL ADR TASKS DROP LDI,      \ Maak adres huidige taak en
    ZH ADR TASKS NIP LDI,
    ZL ADR TASK# ADD,
    CS IF, ZH INC, THEN,
    R16  SPL IN,                \ bewaar RP op data-stack
    -X R16 ST,
    R16  SPH IN,
    -X R16 ST,
    0 [Z] XH ST,                \ en SP in TASKS
    1 [Z] XL ST,
    ADR TASK# INC,              \ Naar volgende taak
    ADR TASK# INC,
    R16 #TASKS 2 * LDI,
    ADR TASK# R16 CP,
    ZER IF,
        ADR TASK# CLR,          \ begin weer bij taak 0
    THEN,
    ZL ADR TASKS DROP LDI,      \ Maak adres volgende taak en
    ZH ADR TASKS NIP LDI,
    ZL ADR TASK# ADD,
    CS IF, ZH INC, THEN,
    XH 0 [Z] LD,                \ haal SP op en
    XL 1 [Z] LD,
    R16 X+ LD,                  \ pop RP van data-stack
    cli,                        \ Interrupts uit
    SPH R16 OUT,
    R16 X+ LD,
    SPL R16 OUT,
    sei,                        \ Interrupts weer aan
    R1 POP,                     \ Herstel loop register
    RET,
END-CODE

\ Hoeft maar een maal uitgevoerd te worden om een taak te initialiseren 
\ en dat moet direct vanuit de basistaak gebeuren !!! (64).
CODE ACTIVATE   ( xt task-nr -- )
    R17 X+ LD,                  \ Task-nr
    R18 X+ LD,                  \ xt-Hi
    R19 X+ LD,                  \ xt-Low
    ZL ADR RETURNSTACKS 4. D- DROP LDI, \ RP basis-adres
    ZH ADR RETURNSTACKS 4. D- NIP LDI,  \ RP basis-adres
    R20 R17 MOV,
    DO,
      ZL RETURNSTACK-SPACE ADIW,    \ Bereken RP positie
    R20 LOOP,
    2 [Z] R18 ST,               \ Plaats xt op R-stack
    3 [Z] R19 ST,
    R18 ZL MOV,                 \ Bewaar RP even
    R19 ZH MOV,
    ZL ADR STACKS 2. D- DROP LDI,   \ SP basis adres
    ZH ADR STACKS 2. D- NIP LDI,    \ SP basis adres
    R20 R17 MOV,
    DO,
      ZL STACK-SPACE ADIW,      \ Bereken SP positie
    R20 LOOP,
    1 [Z] R18 ST,               \ Plaats RP op D-stack
    0 [Z] R19 ST,               \ Plaats RP op D-stack
    R18 ZL MOV,                 \ Bewaar SP even
    R19 ZH MOV,
    R17 LSL,
    ZL ADR TASKS DROP LDI,      \ Bereken TASKS positie
    ZH ADR TASKS NIP LDI,       \ Bereken TASKS positie
    ZL R17 ADD,
    CS IF,  ZH INC,  THEN,
    0 [Z] R19 ST,               \ Bewaar SP in TASKS
    1 [Z] R18 ST,
    RET,
END-CODE


\ Drie afzonderlijke MS timers voor Round Robin voorbeeld

REGISTER TIMER1                         \ Tel milliseconden
REGISTER TIMER2
REGISTER TIMER3

CODE COMPARE-AAN    ( -- )              \ Initialiseer compare interrupt
    R16 $0A LDI,                        \ Gebr. 500 KHz klok, en 'auto clear'
    TCCR1B R16 OUT,
    R16 CLR,                            \ Zet timer 1 registers op nul
    TCNT1H R16 OUT,
    TCNT1L R16 OUT,
    R16 500. NIP LDI,                   \ De cyclus is 1 Ms
    OCR1AH R16 OUT,                     \ Gebruik compare-A registers
    R16 500. DROP LDI,
    OCR1AL R16 OUT,
    R16 $40 LDI,                        \ Timer-1 compare interrupt aan
    TIMSK R16 OUT,
    SEI,
    RET,
END-CODE


\ Interrupt routines

\ Er zijn bij 4 Mhz 4.000.000 tikken per sec. beschikbaar,
\ 1% = 40.000. Drie MS timers duren 1000*26=26.000 ~.65%
CODE TIMER          ( -- )
    R17 PUSH,
    R17 SREG IN,
    adr timer1 tst,
    nzr if,  ADR TIMER1 deC,  then,     \ Werk ms timers bij
    adr timer1 tst,
    nzr if,  ADR TIMER2 deC,  then,
    adr timer1 tst,
    nzr if,  ADR TIMER3 deC,  then,
    SREG R17 OUT,
    R17 POP,
    RETI,
END-CODE  T1-COMPAREA


\ Voor drie afzonderlijke taken een MS routine

: MS1               ( u -- )
    to TIMER1  BEGIN  PAUSE  TIMER1 0= UNTIL
    ;

: MS2               ( u -- )
    to TIMER2  BEGIN  PAUSE  TIMER2 0= UNTIL
    ;

: MS3               ( u -- )
    to TIMER3  BEGIN  PAUSE  TIMER3 0= UNTIL
    ;

                              ( Einde )
