\ ==============================================================
\ [TITLE]     AVR ByteForth pincode lock
\ [FILE]      PINCODE.FRT
\ [CATEGORY]  Pincode lock, 728 bytes
\ [VSN]       3.00
\ [CREATED]   October 5, 1996, Willem Ouwerkerk 
\ [LASTCHNGD] June 30, 2003, Willem Ouwerkerk 
\ [COPYRIGHT] De Elektrise Schaar (C) 1996-2003
\ [USES]      AVR ByteForth cross compiler
\ [PURPOSE]   Reading a key from a 12-keys dump telephone keyboard
\             Connected in a row column matrix. A keyboard like
\             this can be bought for 8,95 at CONRAD electronics.
\ ==============================================================

\G            Keys are connected in a row column matrix, like this:
\G
\G            Keyboard connected to port-1
\G
\G                          1   2   3   4  column number
\G               row mask   1   2   4   8  column mask
\G         row number       |   |   |   |
\G                  0   16--1---4---7---* 
\G                          |   |   |   |
\G                  1   32--2---5---8---0
\G                          |   |   |   |
\G                  2   64--3---6---9---#
\G
\G            How are the bits used:
\G
\G            Bit 0 = Row with 1 2 3 keys
\G            Bit 1 = Row with 4 5 6 keys
\G            Bit 2 = Row with 7 8 9 keys
\G            Bit 3 = Row with * 0 # keys
\G            Bit 4 = Column with 1 4 7 * keys
\G            Bit 5 = Column with 2 5 8 0 keys
\G            Bit 6 = Column with 3 6 9 # keys
\G
\G Principle version of keyboard routine
\G : KEY   ( -- x )        ROW# 4 *  COLUMN# + ;

90S2313                                     \ Setup compiler
2 0 4 MEMORY
NEEDS TARGET

\ Multi level error recovery and execute.
NEEDS CATCH.FRT

ATOM CATCH
ATOM INLINE$

PortB   SFR PORT                            \ Access Port-B
PortD   SFR OUT                             \ Access Port-D
PortD 0 BIT-SFR OUTPUT                      \ Switch output
PortD 2 BIT-SFR NEW-PINCODE-LED             \ Signal enter new pincode
PortD 5 BIT-SFR CONDITION-LED               \ Signal PINCODE condition
                                            \ Flash = Error, 
                                            \ On    = Ready for new input
                                            \ Low   = Wait for key
                                            \ Off   = Lock open

: COLUMN    ( rowmask -- column# )          \ Test keyboard row
    INVERT TO PORT  FROM PORT INVERT 15 AND \ Select row
    4 0 DO  DUP 1 I LSHIFT AND              \ Test a column, key active ?
            IF  DROP I 1+ UNLOOP EXIT  THEN \ Yes, leave column number
    LOOP    DROP FALSE                      \ No key in this row pressed
    ;

: KEYBOARD  ( -- key|quiet )                \ Read key from keyboard
    64 COLUMN ?DUP IF  8 +  EXIT  THEN      \ Row 3
    32 COLUMN ?DUP IF  4 +  EXIT  THEN      \ Row 2
    16 COLUMN                               \ Row 1
    ;

: TRANSLATE ( +n -- char )                  \ Convert keycode to char
    1- S>D  S" 147*2580369#" DROP D+ ROM@   \ for comfort
    ;

: KEY?      ( -- flag ) KEYBOARD 0> ;       \ True if a key is pressed

: KEY       ( -- +n )                       \ Wait for a key press
    BEGIN  
        CLEAR CONDITION-LED  3 MS           \ Short flash of led
        SET CONDITION-LED
        KEY?  12 MS  KEY?  AND              \ Any key pressed
    UNTIL
    KEYBOARD  TRANSLATE                     \ Get key and convert to char
    BEGIN  KEY? 0= UNTIL                    \ Wait for key release
    ;

VARIABLE PINCODE1       VARIABLE PINCODE2   \ Four pincode locations
VARIABLE PINCODE3       VARIABLE PINCODE4

: PINCODE-ERROR ( -- )                      \ Flash condition led for 2 sec.
    20 FOR
        CLEAR CONDITION-LED 100 MS
        SET CONDITION-LED   100 MS
        KEY? IF  UNNEXT EXIT  THEN          \ Quit when key pressed
    NEXT 
    ;

: KEY-OK?       ( #code -- flag )           \ Two keys pressed within limit. ?
    KEY =                                   \ Pincode NOT ok and the next key
    250 FOR                                 \ is pressed within 3 seconds
        CLEAR CONDITION-LED  1 MS           \ Short flash of led
        SET CONDITION-LED
        KEY? ?DUP IF                        \ Key pressed ?
            AND  UNNEXT EXIT                \ Yes, match flags and ready
        THEN  11 MS                         \ No, wait a while
    NEXT
    -1 THROW                                \ Error entering Pincode
    ;

: PINCODE?      ( -- flag )                 \ Leave true if pincode is ok
    FROM PINCODE1 KEY-OK?                   \ Get and test first key
    FROM PINCODE2 KEY-OK?  AND              \ Get and test second key
    FROM PINCODE3 KEY-OK?  AND              \ Get and test third key
    FROM PINCODE4 KEY-OK?  AND              \ Get and test fourth key
    ;

: OPEN-LOCK     ( -- )
    SET CONDITION-LED                       \ Signal lock ativated
    CLEAR OUTPUT                            \ Yes, lock open !
    20 FOR  100 MS  NEXT                    \ Wait 2 seconds
    SET OUTPUT                              \ Close lock again
    CLEAR CONDITION-LED                     \ Signal lock closed
    ;

: SELECT-ACTION ( -- )                      \ Select PINCODE function
    PINCODE? 0= THROW                       \ Ask pincode, stop when error!!
    KEY  DUP [CHAR] * = IF                  \ Enter new pincode ?
        DROP                                \ Yes, do so
        CLEAR NEW-PINCODE-LED               \ Show enter pincode state
        KEY TO PINCODE1  KEY TO PINCODE2    \ Enter four keys
        KEY TO PINCODE3  KEY TO PINCODE4
        SET NEW-PINCODE-LED                 \ Pincode accepted
    ELSE                                    \ No,
        [CHAR] # = IF  OPEN-LOCK  THEN      \ Open lock when only # is
    THEN                                    \ pressed
    ;

: PINCODE       ( -- )                      \ Open lock if pincode ok
    SETUP-BYTEFORTH
    $0F SETDIR PORT  $FF SETDIR OUT         \ Set I/O direction I/O-ports
    SET OUTPUT                              \ Lock closed
    CLEAR CONDITION-LED  250 MS             \ Startup sign
    SET CONDITION-LED                       \ Ready for input
    SET NEW-PINCODE-LED                     \ New pincode state not selected
    &1 TO PINCODE1  &2 TO PINCODE2          \ Default pincode is '1234'
    &3 TO PINCODE3  &4 TO PINCODE4
    BEGIN
        ['] SELECT-ACTION  CATCH            \ Control task with recovery
        IF  PINCODE-ERROR  THEN             \ Show mistake in pincode
    AGAIN
    ;  MAIN

\ End
