\ ==============================================================
\ [TITLE]     AVR ByteForth Hitachi LCD library. 
\ [FILE]      LCD.FRT
\ [CATEGORY]  Library ( 472 bytes )
\ [VSN]       1.12
\ [CREATED]   April 9, 1995, Willem Ouwerkerk 
\ [LASTCHNGD] March 1, 2000, Willem Ouwerkerk 
\ [COPYRIGHT] De Elektrise Schaar (C) 1995-2000
\ [USES]      AVR ByteForth cross compiler
\ [PURPOSE]   LCD-display thru port P1 of the AT89C2051
\             Use a display with the Hitachi HD44780A00H chip,
\             and 4 bits LCD I/O communication (4MHz). 
\ ==============================================================

ATOM INLINE$        ATOM MS             \ Import these macros

\ This method is not valid for a four line UC-16401 display.
\ CR and AT-XY  does not work properly. This because the
\ display has only two lines internally. These lines are
\ interlaced on the four display lines. 

\ LCD-display dimensions and cursor position

40 CONSTANT C/L                         \ Screen width
2 CONSTANT L/S                          \ Lines per screen
VARIABLE OUT                            \ Current cursor position
FLAG DATA?                              \ Lcd data type, true is CHAR

PORTB SFR LCD                           \ I/O port for LCD using 6 bits

\ Send one data nibble (4 bits) x to the LC-display. When flag
\ is true x is accepted as data. When flag is false x is accepted
\ as instruction.
CODE NIBBLE-TO-LCD  ( nibble -- )       \ Lcd
    R16 X+ LD,
    R16 $0F ANDI,                       \ Clear upper four bits
    ADR DATA? BST,
    R16 5 BLD,                          \ Bit P1.5 is RS
    ADR LCD R16 OUT,
    ADR LCD 4 SBI,                      \ Bit P1.4 is E
    ADR LCD 4 CBI,
    RET,
END-CODE

\ Send one character char as two nibbles to a LCD-display. When
\ DATA? is true, char is accepted as data. When DATA? is false,
\ char is accepted as instruction.

\G Stuur het karakter 'char' naar een LC-display.
: LCD-CHAR      ( byte -- )             \ Lcd
    SPLIT  NIBBLE-TO-LCD  NIBBLE-TO-LCD
    50 FOR NEXT                         \ Wait about 40 usec.
    ;

\G Stuur de instructie 'byte' naar een LC-display.
: LCD-INSTR      ( char -- )            \ Lcd "lcd-instruction"
    CLEAR DATA?  LCD-CHAR  SET DATA?  
    ;

\G Maak het LC-display schoon en zet de cursor in linker bovenhoek.
: LCD-PAGE      ( -- )                  \ Lcd Clean LCD and restore cursor
    1 LCD-INSTR  2 MS  CLEAR OUT
    ;

\G Initialiseer een LC-display. Zet 4-bits data bus, maak het 
\G display schoon, zet de cursor uit en zet de display richting 
\G van links naar rechts.
: LCD-INIT      ( -- )                  \ Lcd "lcd-initialise"
    $3F SETDIR LCD                      \ Bit 6 & 7 may be used as inputs
    CLEAR DATA?                         \ Initialise LCD controller
    16 MS                               \ Wait and set 4 bits interface mode
    3 FOR
        %0011 NIBBLE-TO-LCD  5 MS
    NEXT
    %0010 NIBBLE-TO-LCD                 \ Four bits interface set
    $28 LCD-INSTR                       \ Four bits interf. two line display
    $08 LCD-INSTR                       \ Display off
    LCD-PAGE                            \ Clear display
    $06 LCD-INSTR                       \ Shift right
    $0C LCD-INSTR                       \ Display on & cursor off
    ;

\G Zet het karakter 'char' op het LC-display.
: LCD-EMIT      ( char -- )             \ Lcd Output one char
    LCD-CHAR  INCR OUT
    ;

\G Zet 'u' karakters vanaf het ROM adres 'addr' naar het LC-display.
: LCD-TYPE      ( addr u -- )           \ Lcd Output u chars from addr
    ?DUP IF
        FOR  COUNT LCD-EMIT  NEXT  
    THEN
    2DROP
    ;

\G Zet 'u' karakters vanaf het RAM adres 'addr' naar een LC-display.
: LCD-RTYPE     ( addr u -- )           \ Lcd "lcd-ram-type" Output u chars from addr
    ?DUP IF      
        FOR  @+ LCD-EMIT  NEXT  
    THEN
    2DROP
    ;

\G Zet de cursor in de linkerbovenhoek van het LC-display.
: LCD-HOME      ( -- )                  \ Lcd
    $80 LCD-INSTR  CLEAR OUT            \ Put cursor in upper left corner of LCD
    ;

\G Wis het vorige karakter van het LC-display.
: LCD-BS        ( -- )                  \ Lcd "lcd-b-s" Delete previous character if any
    FROM OUT IF
        $10 LCD-INSTR  BL LCD-CHAR
        $10 LCD-INSTR  DECR OUT
    THEN
    ;

\G Zet een spatie op het LC-display.
: LCD-SPACE     ( -- )                  \ Lcd Print one space on a LCD-display
    BL LCD-EMIT
    ;

\G Zet 'u' spaties op het LC-display.
: LCD-SPACES    ( u -- )                \ Lcd Print u spaces on a LCD-display
    ?DUP IF
        FOR  LCD-SPACE  NEXT
    THEN
    ;

\G Zet de cursor op position 'x' 'y' van het LC-display.
\G Let op: Dit werkt zo alleen op 40 kar. displays.
: LCD-AT-XY     ( x y -- )              \ Lcd "lcd-at-xy" Put cursor at pos x of line y
    LCD-HOME  C/L * +  ?DUP IF      
        DUP TO OUT
        FOR  $10 LCD-INSTR  NEXT
    THEN
    ;

\ Does not work for a four line display like UC-16401

\G Doe een CR op een LC-display.
\G Let op: Dit werkt zo alleen op 40 kar. displays.
: LCD-CR        ( -- )                  \ Lcd "lcd-c-r" Advance cursor to a new line
    C/L  FROM OUT  -  ?DUP
    IF  LCD-SPACES  THEN
    FROM OUT  C/L L/S * 1- AND  TO OUT
    ;

\ Generic output routines

MACRO: EMIT      LCD-CHAR ;              ( char -- )
MACRO: TYPE      LCD-TYPE ;              ( addr u -- )
MACRO: RTYPE     LCD-RTYPE ;             ( addr u -- )
MACRO: CR        LCD-CR ;                ( -- )
MACRO: SPACE     LCD-SPACE ;             ( -- )
MACRO: SPACES    LCD-SPACES ;            ( u -- )
MACRO: PAGE      LCD-PAGE ;              ( -- )
MACRO: BS        LCD-BS ;                ( -- )

DOC  LCD demo

: MAIN          ( n -- )                \ Display character set on LCD
    SETUP  LCD-INIT
    BEGIN
        0 FOR
            255 I' - EMIT  50 MS
        NEXT
    AGAIN
    ;   

' MAIN      RESET-VEC  SET-VECTOR

ENDDOC

\ End
